@@ -17,7 +17,7 @@
  *
  *   Copyright © 2008 - 2009 Intel Corporation.
  *
- * Based on nm-openconnect-vpnc.c:
+ * Based on nm-vpnc-service.c:
  *   Copyright © 2005 - 2008 Red Hat, Inc.
  *   Copyright © 2007 - 2008 Novell, Inc.
  */
@@ -40,12 +40,17 @@
 #include <net/if.h>
 #include <pwd.h>
 #include <grp.h>
+#include <locale.h>
 #include <glib/gi18n.h>
 
 #include <nm-setting-vpn.h>
 #include "nm-openconnect-service.h"
 #include "nm-utils.h"
 
+#if !defined(DIST_VERSION)
+# define DIST_VERSION VERSION
+#endif
+
 G_DEFINE_TYPE (NMOPENCONNECTPlugin, nm_openconnect_plugin, NM_TYPE_VPN_PLUGIN)
 
 typedef struct {
@@ -86,6 +91,8 @@ static ValidProperty valid_properties[] = {
 	{ NM_OPENCONNECT_KEY_PROXY,       G_TYPE_STRING, 0, 0 },
 	{ NM_OPENCONNECT_KEY_CSD_ENABLE,  G_TYPE_BOOLEAN, 0, 0 },
 	{ NM_OPENCONNECT_KEY_CSD_WRAPPER, G_TYPE_STRING, 0, 0 },
+	{ NM_OPENCONNECT_KEY_STOKEN_SOURCE, G_TYPE_STRING, 0, 0 },
+	{ NM_OPENCONNECT_KEY_STOKEN_STRING, G_TYPE_STRING, 0, 0 },
 	{ NULL,                           G_TYPE_NONE, 0, 0 }
 };
 
@@ -98,6 +105,8 @@ static ValidProperty valid_secrets[] = {
 
 static uid_t tun_owner;
 static gid_t tun_group;
+static gboolean debug = FALSE;
+static GMainLoop *loop = NULL;
 
 typedef struct ValidateInfo {
 	ValidProperty *table;
@@ -438,6 +447,9 @@ nm_openconnect_start_openconnect_binary (NMOPENCONNECTPlugin *plugin,
 
 	g_ptr_array_add (openconnect_argv, (gpointer) props_vpn_gw);
 
+	if (debug)
+		g_ptr_array_add (openconnect_argv, (gpointer) "--verbose");
+
 	g_ptr_array_add (openconnect_argv, NULL);
 
 	if (!g_spawn_async_with_pipes (NULL, (char **) openconnect_argv->pdata, NULL,
@@ -483,6 +495,9 @@ real_connect (NMVPNPlugin   *plugin,
 	if (!nm_openconnect_secrets_validate (s_vpn, error))
 		goto out;
 
+	if (debug)
+		nm_connection_dump (connection);
+
 	openconnect_fd = nm_openconnect_start_openconnect_binary (NM_OPENCONNECT_PLUGIN (plugin), s_vpn, error);
 	if (!openconnect_fd)
 		return TRUE;
@@ -587,6 +602,27 @@ nm_openconnect_plugin_new (void)
 	                                             NULL);
 }
 
+static void
+signal_handler (int signo)
+{
+	if (signo == SIGINT || signo == SIGTERM)
+		g_main_loop_quit (loop);
+}
+
+static void
+setup_signals (void)
+{
+	struct sigaction action;
+	sigset_t mask;
+
+	sigemptyset (&mask);
+	action.sa_handler = signal_handler;
+	action.sa_mask = mask;
+	action.sa_flags = 0;
+	sigaction (SIGTERM,  &action, NULL);
+	sigaction (SIGINT,  &action, NULL);
+}
+
 static void
 quit_mainloop (NMOPENCONNECTPlugin *plugin, gpointer user_data)
 {
@@ -596,9 +632,45 @@ quit_mainloop (NMOPENCONNECTPlugin *plugin, gpointer user_data)
 int main (int argc, char *argv[])
 {
 	NMOPENCONNECTPlugin *plugin;
-	GMainLoop *main_loop;
 
+	gboolean persist = FALSE;
+	GOptionContext *opt_ctx = NULL;
+
+	GOptionEntry options[] = {
+		{ "persist", 0, 0, G_OPTION_ARG_NONE, &persist, N_("Don't quit when VPN connection terminates"), NULL },
+		{ "debug", 0, 0, G_OPTION_ARG_NONE, &debug, N_("Enable verbose debug logging (may expose passwords)"), NULL },
+		{NULL}
+	};
+
+#if !GLIB_CHECK_VERSION (2, 35, 0)
 	g_type_init ();
+#endif
+
+	/* locale will be set according to environment LC_* variables */
+	setlocale (LC_ALL, "");
+
+	bindtextdomain (GETTEXT_PACKAGE, NM_OPENCONNECT_LOCALEDIR);
+	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
+	textdomain (GETTEXT_PACKAGE);
+
+	/* Parse options */
+	opt_ctx = g_option_context_new (NULL);
+	g_option_context_set_translation_domain (opt_ctx, GETTEXT_PACKAGE);
+	g_option_context_set_ignore_unknown_options (opt_ctx, FALSE);
+	g_option_context_set_help_enabled (opt_ctx, TRUE);
+	g_option_context_add_main_entries (opt_ctx, options, NULL);
+
+	g_option_context_set_summary (opt_ctx,
+		_("nm-openconnect-service provides integrated Cisco AnyConnect SSL VPN capability to NetworkManager."));
+
+	g_option_context_parse (opt_ctx, &argc, &argv, NULL);
+	g_option_context_free (opt_ctx);
+
+	if (getenv ("OPENCONNECT_DEBUG"))
+		debug = TRUE;
+
+	if (debug)
+		g_message ("nm-openconnect-service (version " DIST_VERSION ") starting...");
 
 	if (system ("/sbin/modprobe tun") == -1)
 		exit (EXIT_FAILURE);
@@ -607,15 +679,15 @@ int main (int argc, char *argv[])
 	if (!plugin)
 		exit (EXIT_FAILURE);
 
-	main_loop = g_main_loop_new (NULL, FALSE);
+	loop = g_main_loop_new (NULL, FALSE);
 
-	g_signal_connect (plugin, "quit",
-	                  G_CALLBACK (quit_mainloop),
-	                  main_loop);
+	if (!persist)
+		g_signal_connect (plugin, "quit", G_CALLBACK (quit_mainloop), loop);
 
-	g_main_loop_run (main_loop);
+	setup_signals ();
+	g_main_loop_run (loop);
 
-	g_main_loop_unref (main_loop);
+	g_main_loop_unref (loop);
 	g_object_unref (plugin);
 
 	exit (EXIT_SUCCESS);
